def first(iterable, predicate=lambda x: True, default=None):
    return next((item for item in iterable if predicate(item)), default)


class Immutable:
    """Base immutable class for the query params."""

    _frozen = False

    def __init__(self, mutable=False, **kwargs):
        for kwarg_name, kwarg_value in kwargs.items():
            setattr(self, kwarg_name, kwarg_value)

        self._fields = kwargs
        self._frozen = not mutable

    def __setattr__(self, key, value):
        if self._is_immutable():
            raise TypeError(f"The instance of {self.__class__} is immutable.")
        super().__setattr__(key, value)

    def _is_immutable(self):
        return self._frozen

    def _set_immutable(self):
        """Set the instance as immutable."""
        self._frozen = False

    def as_dict(self):
        """Return the dictionary of items."""
        transformed = {}

        for key, value in self._fields.items():
            if isinstance(value, list):
                value = [
                    x.as_dict() if isinstance(x, self.__class__) else x for x in value
                ]
            elif isinstance(value, self.__class__):
                value = value.as_dict()

            transformed[key] = value

        return transformed



import json
from unittest.mock import Mock

from django.http import QueryDict
from django.test import SimpleTestCase

from mbq.api_tools import fields
from mbq.api_tools.responses import (
    ClientErrorResponse,
    DetailResponse,
    MethodNotAllowedResponse,
    UnauthorizedResponse,
)
from mbq.api_tools.views import View, view


class TruePermissionStub:
    def has_permission(self, request, view):
        return True


class FalsePermissionStub:
    def has_permission(self, request, view):
        return False


class FalsePermissionWithMessageStub:
    message = "Ah ah ah! You didn't say the magic word!"

    def has_permission(self, request, view):
        return False


def view_func(request):
    return DetailResponse()


class ViewFunctionTests(SimpleTestCase):
    def setUp(self):
        super().setUp()
        self.request = Mock(body="", method="GET", GET=QueryDict())

    def test_perform_authorization(self):
        with self.assertRaises(TypeError):
            view(permissions=[])(view_func)

        self.assertTrue(
            isinstance(
                view(permissions=[TruePermissionStub])(view_func)(self.request),
                DetailResponse,
            )
        )

        response = view(permissions=[FalsePermissionStub])(view_func)(self.request)

        self.assertTrue(isinstance(response, UnauthorizedResponse))
        self.assertTrue(response.data["detail"], "Authorization Failed")

        response = view(permissions=[FalsePermissionWithMessageStub])(view_func)(
            self.request
        )

        self.assertTrue(isinstance(response, UnauthorizedResponse))
        self.assertTrue(
            response.data["detail"], "Ah ah ah! You didn't say the magic word!"
        )

        self.assertTrue(
            isinstance(
                view(permissions=[FalsePermissionStub, TruePermissionStub])(view_func)(
                    self.request
                ),
                UnauthorizedResponse,
            )
        )

    def test_method_not_allowed(self):
        self.assertTrue(
            isinstance(
                view("POST", permissions=[TruePermissionStub], payload={})(view_func)(
                    self.request
                ),
                MethodNotAllowedResponse,
            )
        )
        self.assertTrue(
            isinstance(
                view("GET", permissions=[TruePermissionStub])(view_func)(self.request),
                DetailResponse,
            )
        )

    def test_params(self):
        request = Mock(body="", GET=QueryDict("foo=bar&fizz=1"), method="GET")

        def view_func(request):
            return DetailResponse(request.params.as_dict())

        param_schema = {"foo": fields.String(), "fizz": fields.Int()}

        decorated_view = view(permissions=[TruePermissionStub], params=param_schema)(
            view_func
        )

        response = decorated_view(request)

        self.assertEqual(response.data, {"foo": "bar", "fizz": 1})

        request.GET = QueryDict("foo=1")
        response = decorated_view(request)
        self.assertTrue(isinstance(response, ClientErrorResponse))

    def test_payload(self):
        payload = {"foo": "bar", "fizz": 1}

        request = Mock(
            body=json.dumps(payload).encode("utf-8"), GET=QueryDict(), method="POST"
        )

        def view_func(request):
            return DetailResponse(request.payload.as_dict())

        param_schema = {"foo": fields.String(), "fizz": fields.Int()}

        decorated_view = view(
            "POST", permissions=[TruePermissionStub], payload=param_schema
        )(view_func)

        response = decorated_view(request)

        self.assertEqual(response.data, {"foo": "bar", "fizz": 1})

        request.body = json.dumps({"foo": 1}).encode("utf-8")
        response = decorated_view(request)
        self.assertTrue(isinstance(response, ClientErrorResponse))


class TestView(View):
    @view.method("POST", permissions=[TruePermissionStub], payload={})
    def get(self, request):
        return DetailResponse()


class ViewTests(SimpleTestCase):
    def test_method_not_allowed(self):
        request = Mock(method="GET", body="", GET=QueryDict())

        self.assertTrue(
            isinstance(TestView.as_view()(request), MethodNotAllowedResponse)
        )

        request.method = "POST"

        self.assertTrue(isinstance(TestView.as_view()(request), DetailResponse))

    def test_method_must_be_decorated(self):
        class TestView(View):
            def get(self, request):
                return DetailResponse()

        with self.assertRaises(TypeError):
            TestView.as_view()



from django.http import QueryDict
from django.test import SimpleTestCase

from mbq.api_tools import exceptions, fields, schema


class DefaultKwargTests(SimpleTestCase):
    def test_no_default_means_required(self):
        spec = {"name": fields.String()}
        param_schema = schema.make_param_schema(spec)

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load({})

        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()

        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({})

    def test_default_omit(self):
        param_schema = schema.make_param_schema(
            {"name": fields.String(default=fields.OMIT)}
        )

        with self.assertRaises(AttributeError):
            param_schema.load({}).name


class IntFieldTests(SimpleTestCase):
    def test_size(self):
        spec = {"age": fields.Int()}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()

        payload_schema.load({"age": 1})
        payload_schema.load({"age": 7})

        spec = {"age": fields.Int(min_val=5)}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()
        payload_schema.load({"age": 6})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"age": 1})

        spec = {"age": fields.Int(max_val=10)}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()
        payload_schema.load({"age": 7})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"age": 15})

        spec = {"age": fields.Int(min_val=5, max_val=5)}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()
        payload_schema.load({"age": 5})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"age": 1})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"age": 6})


class StringFieldTests(SimpleTestCase):
    def test_missing_value(self):
        param_schema = schema.make_param_schema({"name": fields.String(default=None)})
        result = param_schema.load(QueryDict("name="))
        self.assertEqual(result.name, None)

    def test_allow_empty(self):
        payload_schema = schema.generate_schema(
            schema.PayloadSchema, {"name": fields.String()}
        )()

        result = payload_schema.load({"name": "john"})
        self.assertEqual(result.name, "john")

        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"name": ""})

        payload_schema = schema.generate_schema(
            schema.PayloadSchema, {"name": fields.String(allow_empty=True)}
        )()
        result = payload_schema.load({"name": ""})
        self.assertEqual(result.name, "")

    def test_length(self):
        spec = {"name": fields.String()}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()

        payload_schema.load({"name": "a"})
        payload_schema.load({"name": "abcdefg"})

        spec = {"name": fields.String(min_length=5)}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()
        payload_schema.load({"name": "abcdefg"})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"name": "a"})

        spec = {"name": fields.String(max_length=10)}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()
        payload_schema.load({"name": "abcdefg"})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"name": "abcdefghijklmnop"})

        spec = {"name": fields.String(min_length=5, max_length=5)}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()
        payload_schema.load({"name": "abcde"})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"name": "a"})
        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"name": "abcdefg"})

    def test_min_length_with_allow_empty(self):
        spec = {"name": fields.String(min_length=10, allow_empty=True)}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()

        payload_schema.load({"name": ""})
        payload_schema.load({"name": "abcdefghiklmnop"})

        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"name": "a"})


class EnumFieldTests(SimpleTestCase):
    def test_choices(self):
        spec = {"name": fields.Enum(choices=["foo", "bar"])}
        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()

        payload_schema.load({"name": "foo"})

        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({"name": "fizz"})


class NestedFieldTest(SimpleTestCase):
    def test_basic_param_schema(self):
        """Test that the params schema is generated correctly."""
        param_schema = schema.make_param_schema(
            {
                "category": fields.Nested(
                    {"id": fields.Int(), "description": fields.String()}
                )
            }
        )

        result = param_schema.load(
            {"category": {"id": 42, "description": "Rocket Coffee"}}
        )
        self.assertEqual(result.category.id, 42)
        self.assertEqual(result.category.description, "Rocket Coffee")

    def test_class_param_schema(self):
        class CategorySchema(schema.ParamSchema):
            id = fields.Int()
            type = fields.String()

        param_schema = schema.make_param_schema(
            {"category": fields.Nested(CategorySchema)}
        )

        result = param_schema.load({"category": {"id": 42, "type": "coffee"}})
        self.assertEqual(result.category.id, 42)
        self.assertEqual(result.category.type, "coffee")

    def test_list_param_schema(self):
        """Test that list param schemas are generated correctly."""
        param_schema = schema.make_param_schema(
            {
                "categories": fields.Nested(
                    {"id": fields.Int(), "description": fields.String()}, many=True
                )
            }
        )

        result = param_schema.load(
            {
                "categories": [
                    {"id": 42, "description": "Rocket Coffee"},
                    {"id": 16, "description": "Chocolate Quasar"},
                ]
            }
        )
        self.assertEqual(result.categories[0].id, 42)
        self.assertEqual(result.categories[1].id, 16)

    def test_payload_params(self):
        """Test that nested params works."""
        schema_spec = {
            "offers": fields.Nested({"id": fields.Int()}, many=True),
            "orders": fields.Nested(
                {"id": fields.Int(), "quotes": fields.Int(many=True)}, many=True
            ),
        }
        payload_schema = schema.generate_schema(schema.PayloadSchema, schema_spec)()
        result = payload_schema.load(
            {
                "offers": [{"id": 1}, {"id": 2}],
                "orders": [{"id": 10, "quotes": [1, 2, 3]}],
            }
        )
        self.assertEqual(result.offers[0].id, 1)
        self.assertEqual(len(result.offers), 2)
        self.assertEqual(result.orders[0].id, 10)
        self.assertEqual(len(result.orders[0].quotes), 3)


class DecimalFieldTests(SimpleTestCase):
    def test_validate_field(self):
        param_schema = schema.make_param_schema({"price": fields.Decimal(max_digits=6)})

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("price=1234560"))

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("price=123456.0"))

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("price=12345.60"))

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("price=0.0000001"))


class EmailFieldTests(SimpleTestCase):
    def test_required(self):
        spec = {"email": fields.Email()}

        payload_schema = schema.generate_schema(schema.PayloadSchema, spec)()

        with self.assertRaises(exceptions.ValidationError):
            payload_schema.load({})



import json
from functools import wraps
from typing import Any, Dict, List, Optional, Tuple, Union

from django.conf import settings
from django.views.decorators.csrf import csrf_exempt

from typing_extensions import Literal

from . import exceptions, fields
from .responses import (
    ClientErrorResponse,
    ListResponse,
    MethodNotAllowedResponse,
    PaginatedResponse,
    UnauthorizedResponse,
)
from .schema import Pagination, ParamSchema, PayloadSchema, generate_schema
from .utils import first


PermissionsCollection = Union[List[Any], Tuple[Any, ...]]
HttpMethodNames = Literal["GET", "POST", "PATCH", "PUT", "DELETE"]
OnUnknownField = Literal["raise", "exclude"]


class ViewDecorator:
    def __init__(
        self,
        http_method_name: HttpMethodNames = "GET",
        permissions: Optional[PermissionsCollection] = (),
        params: Optional[Dict[str, fields.Field]] = None,
        payload: Optional[Dict[str, fields.Field]] = None,
        paginated: bool = False,
        page_size: Optional[int] = None,
        on_unknown_field: Optional[OnUnknownField] = None,
        _method: Optional[bool] = None,
    ):
        if not permissions and settings.API_TOOLS.get("REQUIRE_PERMISSIONS", True):
            raise TypeError("Permissions are required")

        if http_method_name in {"POST", "PATCH", "PUT"} and payload is None:
            raise TypeError(
                "A payload must be specified for POST, PATCH, and PUT endpoints (it can be empty)."
            )
        if http_method_name == "GET" and payload is not None:
            raise TypeError("Payloads cannot be specified for GET endpoints")

        self.http_method_name = http_method_name
        self.permissions = permissions
        self.params = params
        self.payload = payload
        self.on_unknown_field = on_unknown_field

        self.paginated = paginated
        self.page_size = page_size

        if self.page_size and not self.paginated:
            raise ValueError(
                f"Defining 'page_size={self.page_size}' requires also setting 'paginated=True'."
            )

        self._method = _method

    def __call__(self, func):
        func.http_method_name = self.http_method_name
        func.permissions = self.permissions
        func.params = self.params
        func.payload = self.payload
        func.paginated = self.paginated
        func.page_size = self.page_size
        func.on_unknown_field = self.on_unknown_field

        if self._method:
            return func

        @wraps(func)
        def wrapped_func(request, *args, **kwargs):
            if request.method.upper() != self.http_method_name:
                return MethodNotAllowedResponse()

            return ViewFunction(func, request)(*args, **kwargs)

        return csrf_exempt(wrapped_func)

    @classmethod
    def method(cls, *args, **kwargs):
        kwargs["_method"] = True
        return cls(*args, **kwargs)


view = ViewDecorator


class View:
    @classmethod
    def as_view(cls):
        view_function_names = [
            attr for attr in dir(cls) if hasattr(getattr(cls, attr), "http_method_name")
        ]

        if not view_function_names:
            raise TypeError(
                "View classes must have at least one method decorated with `view.method`"
            )

        @wraps(cls)
        def dispatcher(request, *args, **kwargs):
            view = cls()
            http_method = request.method.upper()

            view_functions = [
                getattr(view, func_name) for func_name in view_function_names
            ]

            view_method = first(
                view_functions, lambda x: http_method == x.http_method_name
            )

            if view_method is None:
                return MethodNotAllowedResponse()

            return ViewFunction(view_method, request, view_class=cls)(*args, **kwargs)

        return csrf_exempt(dispatcher)


class ViewFunction:
    def __init__(self, view_func, request, view_class=None):
        self.view_func = view_func
        self.request = request
        self.view_class = view_class

        self.permissions = view_func.permissions
        self.params = view_func.params or {}
        self.payload = view_func.payload or {}
        self.paginated = view_func.paginated
        self.page_size = view_func.page_size
        self.on_unknown_field = view_func.on_unknown_field

        # Add the pagination fields to the params
        if self.paginated:
            self.params.setdefault(
                "page", fields.Int(default=Pagination.DEFAULT_PAGE, min_val=0)
            )  # Always page 1
            self.params.setdefault(
                "page_size",
                fields.Int(
                    default=self.page_size or Pagination.DEFAULT_PAGE_SIZE, min_val=1
                ),
            )

    def get_view_name(self):
        func_name = self.view_func.__name__

        if self.view_class:
            return f"{self.view_class.__name__}.{func_name}"
        else:
            return func_name

    def get_param_schema(self):
        schema_class = (
            generate_schema(ParamSchema, self.params) if self.params else ParamSchema
        )
        return schema_class(
            unknown=self.on_unknown_field
            or settings.API_TOOLS.get("UNKNOWN_PARAM_FIELDS", "raise")
        )

    def get_payload_schema(self):
        schema_class = (
            generate_schema(PayloadSchema, self.payload)
            if self.payload
            else PayloadSchema
        )
        return schema_class(
            unknown=self.on_unknown_field
            or settings.API_TOOLS.get("UNKNOWN_PAYLOAD_FIELDS", "raise")
        )

    def __call__(self, *args, **kwargs):
        # Backwards compat thing for DRF permissions
        self.args = args
        self.kwargs = kwargs

        try:
            self._enrich_request()

            authorization_error = self._perform_authorization()
            if authorization_error:
                return UnauthorizedResponse()

            response = self.view_func(self.request, *args, **kwargs)
            self._check_response(response)

            return response

        except exceptions.ValidationError as e:
            return ClientErrorResponse("validation_error", str(e))
        except exceptions.ImmediateResponseError as e:
            return e.response

    def _enrich_request(self):
        payload = (
            json.loads(self.request.body.decode("utf-8")) if self.request.body else {}
        )

        self.request.payload = self.get_payload_schema().load(payload)
        self.request.params = self.get_param_schema().load(self.request.GET)

    def _perform_authorization(self) -> Optional[str]:
        """Returns None if authorization succeeds, or an error message if it fails."""
        permissions = [permission() for permission in self.permissions]

        for permission in permissions:
            if permission.has_permission(self.request, self) is not True:
                # Protecting against mocks with these type checks
                message = getattr(permission, "message", None)
                return message if isinstance(message, str) else "Authorization Failed"

        return None

    def _check_response(self, response):
        """Checks if the response is valid.
        :raises TypeError: If the response type is incorrect
        """
        if (
            response.status_code == 200
            and self.paginated
            and not isinstance(response, PaginatedResponse)
        ):
            raise TypeError(
                f"Invalid pagination, make sure to either set a {PaginatedResponse.__name__} with "
                f"'paginated=True' in the 'view' decorator, or use a {ListResponse.__name__}."
            )



from typing import Optional

from django.db.models import QuerySet
from django.http import JsonResponse

from . import pagination


class BaseResponse(JsonResponse):
    """Base response."""

    def __init__(self, data):
        super().__init__(data, safe=False)


class ListResponse(BaseResponse):
    """List response."""

    status_code = 200

    def __init__(self, objects, meta=None, pagination=None):
        if isinstance(objects, QuerySet):
            objects = list(objects)

        assert not (isinstance(objects, dict) and "objects" in objects)

        self.data = {"objects": objects}
        if meta:
            self.data["meta"] = meta
        if pagination:
            self.data["pagination"] = pagination

        super().__init__(self.data)


class PaginatedResponse(ListResponse):
    """Paginated response."""

    def __init__(self, queryset, serializer, request):
        pagination_params = request.params.pagination

        # Pagination must always exist here
        if not pagination_params:
            raise TypeError(
                f"{PaginatedResponse.__name__} can only be used with a paginated 'view', "
                f"did you mean to use {ListResponse.__name__}?"
            )

        url = request.build_absolute_uri()
        queryset, pagination_data = pagination.paginate_queryset(
            queryset, url, pagination_params
        )

        # If `serializer` is a class, assume it is a DRF serializer.
        if isinstance(serializer, type):
            data = serializer(queryset, many=True).data
        else:
            data = serializer(queryset)

        super().__init__(data, pagination=pagination_data)


class DetailResponse(BaseResponse):
    """Detailed response."""

    status_code = 200

    def __init__(self, object_=""):
        assert not isinstance(object_, list)

        self.data = object_

        super().__init__(object_)


class ErrorResponse(BaseResponse):

    status_code = NotImplemented

    error_code: Optional[str] = None
    detail: Optional[str] = None

    def __init__(
        self, error_code: Optional[str] = None, detail: Optional[str] = None, **kwargs
    ):
        error_code = error_code or self.error_code
        detail = detail or self.detail

        kwargs.update({"error_code": error_code, "detail": detail})
        self.data = kwargs
        super().__init__(self.data)


class ClientErrorResponse(ErrorResponse):
    status_code = 400


class PaginationErrorResponse(ClientErrorResponse):
    error_code = "pagination_error"


class ServerValidationErrorResponse(ErrorResponse):
    status_code = 422


class NotFoundResponse(ErrorResponse):
    status_code = 404
    error_code = "not_found"
    detail = "Resource not found"


class UnauthorizedResponse(ErrorResponse):
    status_code = 403
    error_code = "unauthorized"
    detail = "Unauthorized to access this resource"


class MethodNotAllowedResponse(ErrorResponse):
    status_code = 405
    error_code = "method_not_allowed"
    detail = "This route does not support that HTTP method"



from urllib import parse as urlparse

from django.core.paginator import EmptyPage, Paginator
from django.utils.encoding import force_str

from .exceptions import ImmediateResponseError


def paginate_queryset(queryset, url, pagination_data):
    """Given a queryset and pagination information, return the paginated queryset
    as well as the pagination metadata.

    :param django.Model queryset: The queryset that will be paginated
    :param URL url: A valid url
    :param Pagination pagination: Pagination params
    :return: A tuple containing the paginated query set and the pagination data
    :rtype (queryset, dict)
    """
    paginator = Paginator(queryset, pagination_data.page_size)

    try:
        page = paginator.page(pagination_data.page)
    except EmptyPage:
        from .responses import PaginationErrorResponse

        raise ImmediateResponseError(
            PaginationErrorResponse("That page does not exist")
        )

    return page.object_list, _create_pagination_metadata(page, url)


def _create_pagination_metadata(page, url):
    """Returns a dictionary that contains the pagination metadata
    as well as links to next and previous pages if they are available.

    :param paginator.Page page: The
    :param URL url: A request that knows how to build
     it's own absolute url.
    :return: The pagination dictionary
    :rtype dict
    """
    meta_params = [
        "limit",
        "total_objects",
        "num_pages",
        "page",
        "next_page",
        "previous_page",
    ]
    meta = {param: None for param in meta_params}

    if page is None or not url:
        return meta

    meta["page_size"] = page.paginator.per_page
    meta["page"] = page.number
    meta["next_page"] = _get_page_url(url, page.next_page_number)
    meta["previous_page"] = _get_page_url(url, page.previous_page_number)
    meta["total_objects"] = page.paginator.count

    try:
        meta["num_pages"] = page.paginator.num_pages
    except ZeroDivisionError:
        meta["num_pages"] = 1

    return meta


def _get_page_url(url, page_func):
    """Given a url and a paginator function, return the next/previous page or
    None if the page doesn't exist."""
    try:
        page_number = page_func()
    except (EmptyPage, ZeroDivisionError):
        return None
    return _replace_query_param(url, "page", page_number)


# From rest_framework
def _replace_query_param(url, key, val):
    """
    Given a URL and a key/val pair, set or replace an item in the query
    parameters of the URL, and return the new URL.
    """
    (scheme, netloc, path, query, fragment) = urlparse.urlsplit(force_str(url))
    query_dict = urlparse.parse_qs(query, keep_blank_values=True)
    query_dict[force_str(key)] = [force_str(val)]
    query = urlparse.urlencode(sorted(list(query_dict.items())), doseq=True)
    return urlparse.urlunsplit((scheme, netloc, path, query, fragment))



from datetime import datetime as dt

from django.http import QueryDict
from django.test import SimpleTestCase

from dateutil.tz import tzutc

from mbq.api_tools import exceptions, fields
from mbq.api_tools.schema import ParamSchema, generate_schema


def _make_schema(spec):
    return generate_schema(ParamSchema, spec)()


class SerializeCustomQueryParamsTests(SimpleTestCase):
    def test_parse_single_query_param(self):
        data = _make_schema({"company": fields.Int(param_name="company_id")}).load(
            QueryDict("company_id=19")
        )
        self.assertEqual(data.company, 19)

    def test_parse_multiple_query_params(self):
        mock_params = QueryDict(
            "company_id=19"
            "&telephone=132-4213-123321"
            "&recurring=true"
            "&start_date=2017-02-03T22:00:00Z"
        )

        param_schema = _make_schema(
            {
                "company": fields.Int(param_name="company_id"),
                "telephone": fields.String(),
                "is_recurring": fields.Bool(param_name="recurring"),
                "start_date": fields.DateTime(),
            }
        )
        data = param_schema.load(mock_params)

        self.assertEqual(data.company, 19)
        self.assertEqual(data.telephone, "132-4213-123321")
        self.assertEqual(data.is_recurring, True)
        self.assertEqual(data.start_date, dt(2017, 2, 3, 22, tzinfo=tzutc()))

    def test_parse_query_params_with_defaults(self):
        param_schema = _make_schema(
            {
                "company_id": fields.Int(param_name="company_id", default=1),
                "is_recurring": fields.Bool(param_name="recurring", default=False),
                "start_date": fields.DateTime(
                    param_name="start_date", default=dt(1999, 8, 12, 4, tzinfo=tzutc())
                ),
            }
        )
        data = param_schema.load(QueryDict())

        self.assertEqual(data.company_id, 1)
        self.assertEqual(data.is_recurring, False)
        self.assertEqual(data.start_date, dt(1999, 8, 12, 4, tzinfo=tzutc()))

    def test_parse_query_params_required(self):
        param_schema = _make_schema({"company_id": fields.Int(param_name="company_id")})
        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict())

    def test_parse_query_params_validate(self):
        mock_params = QueryDict("price=1000")

        param_schema = _make_schema(
            {"price": fields.Int(param_name="price", validate=lambda n: n < 100)}
        )
        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(mock_params)

    def test_datetime_query_param(self):
        param_schema = _make_schema({"foo": fields.DateTime()})

        # Fails if there's no timezone
        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("foo=2017-02-03T22:00:0"))

    def test_invalid_int_query_param(self):
        param_schema = _make_schema({"foo": fields.Int()})

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("foo=bar"))

    def test_invalid_bool_query_param(self):
        param_schema = _make_schema({"foo": fields.Bool()})

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("foo=bar"))

    def test_invalid_date_query_param(self):
        param_schema = _make_schema({"foo": fields.DateTime()})

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("foo=bar"))


class SerializeListQueryParamTests(SimpleTestCase):
    def test_basic_input(self):
        """Basic/happy path tests."""
        mock_params = QueryDict("ids=1,2,3")

        param_schema = _make_schema({"ids": fields.Int(many=True)})
        query_params = param_schema.load(mock_params)

        self.assertEqual(query_params.ids, [1, 2, 3])

    def test_handles_different_inputs(self):
        """Make sure that we handle csv and standard list inputs."""
        mock_params = QueryDict("building_ids=one,two,three&ids=1&ids=2&ids=3")

        param_schema = _make_schema(
            {"building_ids": fields.String(many=True), "ids": fields.Int(many=True)}
        )
        query_params = param_schema.load(mock_params)

        self.assertEqual(query_params.building_ids, ["one", "two", "three"])
        self.assertEqual(query_params.ids, [1, 2, 3])

    def test_query_params_types(self):
        """Make sure it can handle any of the basic types."""
        mock_params = QueryDict(
            "ids=1&ids=42&ids=158"
            "&person_id=123&person_id=456&person_id=789"
            "&validated=false&validated=true"
            "&start_date=2017-02-03T22:00:00Z&start_date=1987-11-24T23:50:00Z"
        )

        param_schema = _make_schema(
            {
                "company_ids": fields.Int(many=True, param_name="ids"),
                "person_ids": fields.String(many=True, param_name="person_id"),
                "is_validated": fields.Bool(many=True, param_name="validated"),
                "start_dates": fields.DateTime(many=True, param_name="start_date"),
            }
        )
        query_params = param_schema.load(mock_params)

        def assert_type(value_type, values):
            return all(type(value) == value_type for value in values)

        self.assertEqual(query_params.company_ids, [1, 42, 158])
        self.assertEqual(query_params.person_ids, ["123", "456", "789"])
        self.assertEqual(query_params.is_validated, [False, True])
        self.assertEqual(
            query_params.start_dates,
            [
                dt(2017, 2, 3, 22, tzinfo=tzutc()),
                dt(1987, 11, 24, 23, 50, tzinfo=tzutc()),
            ],
        )

        assert_type(int, query_params.company_ids)
        assert_type(str, query_params.person_ids)
        assert_type(bool, query_params.is_validated)
        assert_type(dt, query_params.start_dates)

    def test_handles_default_values(self):
        param_schema = _make_schema(
            {"ids": fields.Int(many=True, param_name="buildings", default=[1, 2])}
        )
        query_params = param_schema.load(QueryDict())

        self.assertEqual(query_params.ids, [1, 2])

    def test_required_raises(self):
        """Handles that marking as required will correctly raise."""

        param_schema = _make_schema({"ids": fields.Int(many=True)})

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict())

    def test_validation_raises(self):
        """Make sure the correct exception raises if validation fails."""
        VALID_STATUSES = ["active", "approved"]

        def validate_status(statuses):
            if not all(status in VALID_STATUSES for status in statuses):
                raise exceptions.ValidationError("Invalid status!")

        param_schema = _make_schema(
            {"partner_status": fields.String(many=True, validate=validate_status)}
        )

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("partner_status=registered"))

    def test_unserializable_value_raises(self):
        """Unserializable inputs should always raise"""
        param_schema = _make_schema({"ids": fields.Int(many=True)})

        with self.assertRaises(exceptions.ValidationError):
            param_schema.load(QueryDict("ids=one,two,three"))


class SerializePaginatedQueryParamsTests(SimpleTestCase):
    def test_simple_pagination(self):
        param_schema = _make_schema({"page": fields.Int(), "page_size": fields.Int()})
        pagination = param_schema.load(QueryDict("page=1&page_size=10")).pagination

        self.assertEqual(pagination.page, 1)
        self.assertEqual(pagination.page_size, 10)

    def test_default_pagination(self):
        param_schema = _make_schema(
            {"page": fields.Int(default=2), "page_size": fields.Int(default=25)}
        )
        pagination = param_schema.load(QueryDict()).pagination

        self.assertEqual(pagination.page, 2)
        self.assertEqual(pagination.page_size, 25)



from copy import copy

from django.conf import settings

from marshmallow import Schema as MarshmallowSchema
from marshmallow import ValidationError as MarshmallowValidationError
from marshmallow import post_load, pre_load

from . import exceptions, fields, utils


# Responses
class Pagination(utils.Immutable):
    """Immutable pagination query parameters."""

    DEFAULT_PAGE = 1
    DEFAULT_PAGE_SIZE = settings.API_TOOLS.get("DEFAULT_PAGE_SIZE", 20)


class Params(utils.Immutable):
    """Immutable request query parameters."""

    def __init__(self, **kwargs):
        pagination_data = {}

        if "page" in kwargs and "page_size" in kwargs:
            pagination_data["page"] = kwargs.pop("page", None)
            pagination_data["page_size"] = kwargs.pop("page_size", None)

        self.pagination = Pagination(**pagination_data) if pagination_data else None
        super().__init__(**kwargs)


class Payload(utils.Immutable):
    """Immutable payload data."""


def generate_schema(schema_class, spec):
    spec = copy(spec)
    spec["original_schema_class"] = schema_class
    return type("GeneratedSchema", (schema_class,), spec)


def make_param_schema(spec):
    return generate_schema(ParamSchema, spec)()


class Schema(MarshmallowSchema):
    """."""

    def __init__(self, unknown=None, **kwargs):
        allowed_kwargs = ("many",)
        filtered_kwargs = {
            kwarg: value for kwarg, value in kwargs.items() if kwarg in allowed_kwargs
        }
        super().__init__(unknown=unknown, **filtered_kwargs)

    def load(self, data):
        """Overwrite any `ValidationError`s that may have been risen throughout the load
        process and re-raise it as our own custom error.

        :param dict data: The data that will get deserialized
        :rtype utils.Immutable
        :returns An immutable class with the query params
        """
        try:
            return super().load(data)
        except MarshmallowValidationError as err:
            raise exceptions.ValidationError(err)

    @pre_load
    def call_pre_load_processes(self, data, *args, **kwargs):
        self._generate_nested_fields()
        return self.pre_process_data(data)

    @post_load
    def set_full_result(self, result, *args, **kwargs):
        return self.output_cls(**result)

    def pre_process_data(self, data):
        """To be overridden by subclasses."""
        return data

    def _generate_nested_fields(self):
        for _, field_obj in self.fields.items():
            if isinstance(field_obj, fields.Nested) and field_obj.needs_schema():
                nested = generate_schema(self.original_schema_class, field_obj.fields)()
                field_obj.bind_schema(nested)


class ParamSchema(Schema):
    """Custom param schema to automatically parse query params from a request
        whilst enabling all of the validation features available in marshmallow.

        Example usage::

        .. code-block:: python
            class QueryParam(ParamSchema):
                order_ids = ListParam(StringParam(), param_name='ids')
                company = StringParam(param_name='company_id')

                @validates(company)
                def validate_company_id(id):
                    if not id:
                        raise ValidationError('The company id cannot be empty')

            order_query_params = QueryParam()
            data = order_query_params.load(request)
            data # {'order_ids': [...], 'company': '..'}
    """

    output_cls = Params

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def pre_process_data(self, data):
        return self._transform_from_querydict(data)

    def _transform_from_querydict(self, data):
        # Fields can be declared with a `param_name` different than the attribute name.
        # Map what the expected param_name is to the field_obj.
        declared_fields = {
            (field_obj.data_key or field): field_obj
            for field, field_obj in self._declared_fields.items()
        }
        transformed = {}

        for key, value in data.items():

            if isinstance(declared_fields.get(key), fields._List):
                value = data.getlist(key)
                # If the param was passed, but no value specified, we get empty strings, which we
                # want to prune.
                value = [x for x in value if x != ""]

                # getlist will return a list containing a single string if the param is given as
                # a comma separated value (ex. 'foo=1,2,3'). We need to split the value in that case.
                if len(value) == 1:
                    value = value[0].split(",")

            # If a param is passed in with no value, drop it.
            if value in ([], ""):
                continue

            transformed[key] = value

        return transformed


class PayloadSchema(Schema):

    output_cls = Payload
